#include "config.h"

#include "game_str.h"

/* -------------------------------------------------------------------------- */

const char *game_str_tbl_ship_names[RACE_NUM * SHIP_NAME_NUM] = {
    "FIGHTER", "GUNBOAT", "COURIER", "FRIGATE", "DESTROYER", "CORVETTE", "CRUISER", "ESCORT", "WARSHIP", "DREADNOUGHT", "BATTLESHIP", "DREADSTAR",
    "WOLVERINE", "GRIFFIN", "FERRET", "LYNX", "CHEETAH", "BOBCAT", "PANTHER", "LEOPARD", "WARCAT", "SABRETOOTH", "PUMA", "TIGER",
    "PIRANHA", "STINGRAY", "MINNOW", "MANTA", "BARRACUDA", "MOREY", "SHARK", "MAKO", "WHALE", "KRAKEN", "MONITOR", "POLARIS",
    "GOBLIN", "CYCLOPS", "DAEMON", "BANSHEE", "SPECTRE", "SPIRIT", "JUGGERNAUT", "VALKYRIE", "HYDRA", "TITAN", "COLOSSUS", "DRAGON",
    "STARFIGHTER", "COMET", "SEEKER", "STAR WING", "STAR BLAZER", "STAR STREAK", "STAR BLADE", "DARK STAR", "CORAONA", "SUN FIRE", "STAR FURY", "NOVA",
    "FOXBAT", "SPARROWHAWK", "PELICAN", "FALCON", "SKY HAWK", "SPACE GULL", "WARBIRD", "WARHAWK", "WAREAGLE", "CONDOR", "DREADWING", "SKY MASTER",
    "SABRE", "CUTLASS", "DAGGER", "LANCER", "PEGASUS", "HORSEMAN", "RANGER", "KNIGHT", "AVENGER", "DRAGOON", "PALADIN", "EXCALIBER",
    "PAW", "PANDA", "PATROL", "HUNTER", "GLADIATOR", "SENTINEL", "CLAW", "TOOTH", "WARBEAR", "GRIZZLY", "PUNISHER", "RHINO",
    "MARAUDER", "HYPERION", "DRACHMA", "INTRUDER", "PENETRATOR", "TORMENTER", "AJAX", "TORNADO", "NEXUS", "NEMESIS", "DEVASTATOR", "ANNIHILATOR",
    "SPIDER", "BUMBLEBEE", "NEEDLE", "HORNET", "SCORPION", "VENOM", "TARANTULA", "VIPER", "COBRA", "KING COBRA", "BLACK WIDOW", "DEATH WYN"
};
