#ifndef INC_1OOM_GAME_TECHTYPES_H
#define INC_1OOM_GAME_TECHTYPES_H

typedef enum {
    TECH_COMP_NONE = 0,
    TECH_COMP_BATTLE_SCANNER, /*1*/
    TECH_COMP_ECM_JAMMER_MARK_I, /*2*/
    TECH_COMP_UNUSED_2, /*3*/
    TECH_COMP_DEEP_SPACE_SCANNER, /*4*/
    TECH_COMP_BATTLE_COMPUTER_MARK_II, /*5*/
    TECH_COMP_UNUSED_5, /*6*/
    TECH_COMP_ECM_JAMMER_MARK_II, /*7*/
    TECH_COMP_IMPROVED_ROBOTIC_CONTROLS_III, /*8*/
    TECH_COMP_UNUSED_8, /*9*/
    TECH_COMP_BATTLE_COMPUTER_MARK_III, /*10*/
    TECH_COMP_UNUSED_10, /*11*/
    TECH_COMP_ECM_JAMMER_MARK_III, /*12*/
    TECH_COMP_IMPROVED_SPACE_SCANNER, /*13*/
    TECH_COMP_UNUSED_13, /*14*/
    TECH_COMP_BATTLE_COMPUTER_MARK_IV, /*15*/
    TECH_COMP_UNUSED_15, /*16*/
    TECH_COMP_ECM_JAMMER_MARK_IV, /*17*/
    TECH_COMP_IMPROVED_ROBOTIC_CONTROLS_IV, /*18*/
    TECH_COMP_UNUSED_18, /*19*/
    TECH_COMP_BATTLE_COMPUTER_MARK_V, /*20*/
    TECH_COMP_UNUSED_20, /*21*/
    TECH_COMP_ECM_JAMMER_MARK_V, /*22*/
    TECH_COMP_ADVANCED_SPACE_SCANNER, /*23*/
    TECH_COMP_UNUSED_23, /*24*/
    TECH_COMP_BATTLE_COMPUTER_MARK_VI, /*25*/
    TECH_COMP_UNUSED_25, /*26*/
    TECH_COMP_ECM_JAMMER_MARK_VI, /*27*/
    TECH_COMP_IMPROVED_ROBOTIC_CONTROLS_V, /*28*/
    TECH_COMP_UNUSED_28, /*29*/
    TECH_COMP_BATTLE_COMPUTER_MARK_VII, /*30*/
    TECH_COMP_UNUSED_30, /*31*/
    TECH_COMP_ECM_JAMMER_MARK_VII, /*32*/
    TECH_COMP_UNUSED_32, /*33*/
    TECH_COMP_HYPERSPACE_COMMUNICATIONS, /*34*/
    TECH_COMP_BATTLE_COMPUTER_MARK_VIII, /*35*/
    TECH_COMP_UNUSED_35, /*36*/
    TECH_COMP_ECM_JAMMER_MARK_VIII, /*37*/
    TECH_COMP_IMPROVED_ROBOTIC_CONTROLS_VI, /*38*/
    TECH_COMP_UNUSED_38, /*39*/
    TECH_COMP_BATTLE_COMPUTER_MARK_IX, /*40*/
    TECH_COMP_UNUSED_40, /*41*/
    TECH_COMP_ECM_JAMMER_MARK_IX, /*42*/
    TECH_COMP_UNUSED_42, /*43*/
    TECH_COMP_UNUSED_43, /*44*/
    TECH_COMP_BATTLE_COMPUTER_MARK_X, /*45*/
    TECH_COMP_ORACLE_INTERFACE, /*46*/
    TECH_COMP_ECM_JAMMER_MARK_X, /*47*/
    TECH_COMP_IMPROVED_ROBOTIC_CONTROLS_VII, /*48*/
    TECH_COMP_TECHNOLOGY_NULLIFIER, /*49*/
    TECH_COMP_BATTLE_COMPUTER_MARK_XI /*50*/
} tech_comp_t;

typedef enum {
    TECH_CONS_NONE = 0,
    TECH_CONS_RESERVE_FUEL_TANKS, /*1*/
    TECH_CONS_UNUSED_1, /*2*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_9, /*3*/
    TECH_CONS_UNUSED_3, /*4*/
    TECH_CONS_REDUCED_INDUSTRIAL_WASTE_80, /*5*/
    TECH_CONS_UNUSED_5, /*6*/
    TECH_CONS_UNUSED_6, /*7*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_8, /*8*/
    TECH_CONS_UNUSED_8, /*9*/
    TECH_CONS_DURALLOY_ARMOR, /*10*/
    TECH_CONS_BATTLE_SUITS, /*11*/
    TECH_CONS_UNUSED_11, /*12*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_7, /*13*/
    TECH_CONS_AUTOMATED_REPAIR_SYSTEM, /*14*/
    TECH_CONS_REDUCED_INDUSTRIAL_WASTE_60, /*15*/
    TECH_CONS_UNUSED_15, /*16*/
    TECH_CONS_ZORTIUM_ARMOR, /*17*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_6, /*18*/
    TECH_CONS_UNUSED_18, /*19*/
    TECH_CONS_UNUSED_19, /*20*/
    TECH_CONS_UNUSED_20, /*21*/
    TECH_CONS_UNUSED_21, /*22*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_5, /*23*/
    TECH_CONS_ARMORED_EXOSKELETON, /*24*/
    TECH_CONS_REDUCED_INDUSTRIAL_WASTE_40, /*25*/
    TECH_CONS_ANDRIUM_ARMOR, /*26*/
    TECH_CONS_UNUSED_26, /*27*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_4, /*28*/
    TECH_CONS_UNUSED_28, /*29*/
    TECH_CONS_UNUSED_29, /*30*/
    TECH_CONS_UNUSED_30, /*31*/
    TECH_CONS_UNUSED_31, /*32*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_3, /*33*/
    TECH_CONS_TRITANIUM_ARMOR, /*34*/
    TECH_CONS_REDUCED_INDUSTRIAL_WASTE_20, /*35*/
    TECH_CONS_ADVANCED_DAMAGE_CONTROL, /*36*/
    TECH_CONS_UNUSED_36, /*37*/
    TECH_CONS_IMPROVED_INDUSTRIAL_TECH_2, /*38*/
    TECH_CONS_UNUSED_38, /*39*/
    TECH_CONS_POWERED_ARMOR, /*40*/
    TECH_CONS_UNUSED_40, /*41*/
    TECH_CONS_ADAMANTIUM_ARMOR, /*42*/
    TECH_CONS_UNUSED_42, /*43*/
    TECH_CONS_UNUSED_43, /*44*/
    TECH_CONS_INDUSTRIAL_WASTE_ELIMINATION, /*45*/
    TECH_CONS_UNUSED_45, /*46*/
    TECH_CONS_UNUSED_46, /*47*/
    TECH_CONS_UNUSED_47, /*48*/
    TECH_CONS_UNUSED_48, /*49*/
    TECH_CONS_NEUTRONIUM_ARMOR /*50*/
} tech_cons_t;

typedef enum {
    TECH_FFLD_NONE = 0,
    TECH_FFLD_CLASS_I_DEFLECTOR_SHIELDS, /*1*/
    TECH_FFLD_UNUSED_1, /*2*/
    TECH_FFLD_UNUSED_2, /*3*/
    TECH_FFLD_CLASS_II_DEFLECTOR_SHIELDS, /*4*/
    TECH_FFLD_UNUSED_4, /*5*/
    TECH_FFLD_UNUSED_5, /*6*/
    TECH_FFLD_UNUSED_6, /*7*/
    TECH_FFLD_PERSONAL_DEFLECTOR_SHIELD, /*8*/
    TECH_FFLD_UNUSED_8, /*9*/
    TECH_FFLD_CLASS_III_DEFLECTOR_SHIELDS, /*10*/
    TECH_FFLD_UNUSED_10, /*11*/
    TECH_FFLD_CLASS_V_PLANETARY_SHIELD, /*12*/
    TECH_FFLD_UNUSED_12, /*13*/
    TECH_FFLD_CLASS_IV_DEFLECTOR_SHIELDS, /*14*/
    TECH_FFLD_UNUSED_14, /*15*/
    TECH_FFLD_REPULSOR_BEAM, /*16*/
    TECH_FFLD_UNUSED_16, /*17*/
    TECH_FFLD_UNUSED_17, /*18*/
    TECH_FFLD_UNUSED_18, /*19*/
    TECH_FFLD_CLASS_V_DEFLECTOR_SHIELDS, /*20*/
    TECH_FFLD_PERSONAL_ABSORPTION_SHIELD, /*21*/
    TECH_FFLD_CLASS_X_PLANETARY_SHIELD, /*22*/
    TECH_FFLD_UNUSED_22, /*23*/
    TECH_FFLD_CLASS_VI_DEFLECTOR_SHIELDS, /*24*/
    TECH_FFLD_UNUSED_24, /*25*/
    TECH_FFLD_UNUSED_25, /*26*/
    TECH_FFLD_CLOAKING_DEVICE, /*27*/
    TECH_FFLD_UNUSED_27, /*28*/
    TECH_FFLD_UNUSED_28, /*29*/
    TECH_FFLD_CLASS_VII_DEFLECTOR_SHIELDS, /*30*/
    TECH_FFLD_ZYRO_SHIELD, /*31*/
    TECH_FFLD_CLASS_XV_PLANETARY_SHIELD, /*32*/
    TECH_FFLD_UNUSED_32, /*33*/
    TECH_FFLD_CLASS_IX_DEFLECTOR_SHIELDS, /*34*/
    TECH_FFLD_UNUSED_34, /*35*/
    TECH_FFLD_UNUSED_35, /*36*/
    TECH_FFLD_STASIS_FIELD, /*37*/
    TECH_FFLD_PERSONAL_BARRIER_SHIELD, /*38*/
    TECH_FFLD_UNUSED_38, /*39*/
    TECH_FFLD_CLASS_XI_DEFLECTOR_SHIELDS, /*40*/
    TECH_FFLD_UNUSED_40, /*41*/
    TECH_FFLD_CLASS_XX_PLANETARY_SHIELD, /*42*/
    TECH_FFLD_BLACK_HOLE_GENERATOR, /*43*/
    TECH_FFLD_CLASS_XIII_DEFLECTOR_SHIELDS, /*44*/
    TECH_FFLD_UNUSED_44, /*45*/
    TECH_FFLD_LIGHTNING_SHIELD, /*46*/
    TECH_FFLD_UNUSED_46, /*47*/
    TECH_FFLD_UNUSED_47, /*48*/
    TECH_FFLD_UNUSED_48, /*49*/
    TECH_FFLD_CLASS_XV_DEFLECTOR_SHIELDS /*50*/
} tech_ffld_t;

typedef enum {
    TECH_PLAN_NONE = 0,
    TECH_PLAN_ECOLOGICAL_RESTORATION, /*1*/
    TECH_PLAN_IMPROVED_TERRAFORMING__10, /*2*/
    TECH_PLAN_CONTROLLED_BARREN_ENVIRONMENT, /*3*/
    TECH_PLAN_UNUSED_3, /*4*/
    TECH_PLAN_IMPROVED_ECO_RESTORATION, /*5*/
    TECH_PLAN_CONTROLLED_TUNDRA_ENVIRONMENT, /*6*/
    TECH_PLAN_UNUSED_6, /*7*/
    TECH_PLAN_IMPROVED_TERRAFORMING__20, /*8*/
    TECH_PLAN_CONTROLLED_DEAD_ENVIRONMENT, /*9*/
    TECH_PLAN_DEATH_SPORES, /*10*/
    TECH_PLAN_UNUSED_10, /*11*/
    TECH_PLAN_CONTROLLED_INFERNO_ENVIRONMENT, /*12*/
    TECH_PLAN_ENHANCED_ECO_RESTORATION, /*13*/
    TECH_PLAN_IMPROVED_TERRAFORMING__30, /*14*/
    TECH_PLAN_CONTROLLED_TOXIC_ENVIRONMENT, /*15*/
    TECH_PLAN_SOIL_ENRICHMENT, /*16*/
    TECH_PLAN_BIO_TOXIN_ANTIDOTE, /*17*/
    TECH_PLAN_CONTROLLED_RADIATED_ENVIRONMENT, /*18*/
    TECH_PLAN_UNUSED_18, /*19*/
    TECH_PLAN_IMPROVED_TERRAFORMING__40, /*20*/
    TECH_PLAN_CLONING, /*21*/
    TECH_PLAN_ATMOSPHERIC_TERRAFORMING, /*22*/
    TECH_PLAN_UNUSED_22, /*23*/
    TECH_PLAN_ADVANCED_ECO_RESTORATION, /*24*/
    TECH_PLAN_UNUSED_24, /*25*/
    TECH_PLAN_IMPROVED_TERRAFORMING__50, /*26*/
    TECH_PLAN_DOOM_VIRUS, /*27*/
    TECH_PLAN_UNUSED_27, /*28*/
    TECH_PLAN_UNUSED_28, /*29*/
    TECH_PLAN_ADVANCED_SOIL_ENRICHMENT, /*30*/
    TECH_PLAN_UNUSED_30, /*31*/
    TECH_PLAN_IMPROVED_TERRAFORMING__60, /*32*/
    TECH_PLAN_UNUSED_32, /*33*/
    TECH_PLAN_COMPLETE_ECO_RESTORATION, /*34*/
    TECH_PLAN_UNUSED_34, /*35*/
    TECH_PLAN_UNIVERSAL_ANTIDOTE, /*36*/
    TECH_PLAN_UNUSED_36, /*37*/
    TECH_PLAN_IMPROVED_TERRAFORMING__80, /*38*/
    TECH_PLAN_UNUSED_38, /*39*/
    TECH_PLAN_BIO_TERMINATOR, /*40*/
    TECH_PLAN_UNUSED_40, /*41*/
    TECH_PLAN_ADVANCED_CLONING, /*42*/
    TECH_PLAN_UNUSED_42, /*43*/
    TECH_PLAN_IMPROVED_TERRAFORMING__100, /*44*/
    TECH_PLAN_UNUSED_44, /*45*/
    TECH_PLAN_UNUSED_45, /*46*/
    TECH_PLAN_UNUSED_46, /*47*/
    TECH_PLAN_UNUSED_47, /*48*/
    TECH_PLAN_UNUSED_48, /*49*/
    TECH_PLAN_COMPLETE_TERRAFORMING /*50*/
} tech_plan_t;

typedef enum {
    TECH_PROP_NONE = 0,
    TECH_PROP_RETRO_ENGINES, /*1*/
    TECH_PROP_UNUSED_1, /*2*/
    TECH_PROP_HYDROGEN_FUEL_CELLS, /*3*/
    TECH_PROP_UNUSED_3, /*4*/
    TECH_PROP_DEUTERIUM_FUEL_CELLS, /*5*/
    TECH_PROP_NUCLEAR_ENGINES, /*6*/
    TECH_PROP_UNUSED_6, /*7*/
    TECH_PROP_UNUSED_7, /*8*/
    TECH_PROP_IRRIDIUM_FUEL_CELLS, /*9*/
    TECH_PROP_INERTIAL_STABILIZER, /*10*/
    TECH_PROP_UNUSED_10, /*11*/
    TECH_PROP_SUB_LIGHT_DRIVES, /*12*/
    TECH_PROP_UNUSED_12, /*13*/
    TECH_PROP_DOTOMITE_CRYSTALS, /*14*/
    TECH_PROP_UNUSED_14, /*15*/
    TECH_PROP_ENERGY_PULSAR, /*16*/
    TECH_PROP_UNUSED_16, /*17*/
    TECH_PROP_FUSION_DRIVES, /*18*/
    TECH_PROP_URIDIUM_FUEL_CELLS, /*19*/
    TECH_PROP_WARP_DISSIPATOR, /*20*/
    TECH_PROP_UNUSED_20, /*21*/
    TECH_PROP_UNUSED_21, /*22*/
    TECH_PROP_REAJAX_II_FUEL_CELLS, /*23*/
    TECH_PROP_IMPULSE_DRIVES, /*24*/
    TECH_PROP_UNUSED_24, /*25*/
    TECH_PROP_UNUSED_25, /*26*/
    TECH_PROP_INTERGALACTIC_STAR_GATES, /*27*/
    TECH_PROP_UNUSED_27, /*28*/
    TECH_PROP_TRILITHIUM_CRYSTALS, /*29*/
    TECH_PROP_ION_DRIVES, /*30*/
    TECH_PROP_UNUSED_30, /*31*/
    TECH_PROP_UNUSED_31, /*32*/
    TECH_PROP_UNUSED_32, /*33*/
    TECH_PROP_HIGH_ENERGY_FOCUS, /*34*/
    TECH_PROP_UNUSED_34, /*35*/
    TECH_PROP_ANTI_MATTER_DRIVES, /*36*/
    TECH_PROP_UNUSED_36, /*37*/
    TECH_PROP_SUB_SPACE_TELEPORTER, /*38*/
    TECH_PROP_UNUSED_38, /*39*/
    TECH_PROP_IONIC_PULSAR, /*40*/
    TECH_PROP_THORIUM_CELLS, /*41*/
    TECH_PROP_INTER_PHASED_DRIVES, /*42*/
    TECH_PROP_SUB_SPACE_INTERDICTOR, /*43*/
    TECH_PROP_UNUSED_43, /*44*/
    TECH_PROP_COMBAT_TRANSPORTERS, /*45*/
    TECH_PROP_INERTIAL_NULLIFIER, /*46*/
    TECH_PROP_UNUSED_46, /*47*/
    TECH_PROP_HYPER_DRIVES, /*48*/
    TECH_PROP_UNUSED_48, /*49*/
    TECH_PROP_DISPLACEMENT_DEVICE /*50*/
} tech_prop_t;

typedef enum {
    TECH_WEAP_NONE = 0,
    TECH_WEAP_LASERS, /*1*/
    TECH_WEAP_HAND_LASERS, /*2*/
    TECH_WEAP_UNUSED_2, /*3*/
    TECH_WEAP_HYPER_V_ROCKETS, /*4*/
    TECH_WEAP_GATLING_LASER, /*5*/
    TECH_WEAP_ANTI_MISSILE_ROCKETS, /*6*/
    TECH_WEAP_NEUTRON_PELLET_GUN, /*7*/
    TECH_WEAP_HYPER_X_ROCKETS, /*8*/
    TECH_WEAP_FUSION_BOMB, /*9*/
    TECH_WEAP_ION_CANNON, /*10*/
    TECH_WEAP_SCATTER_PACK_V_ROCKETS, /*11*/
    TECH_WEAP_ION_RIFLE, /*12*/
    TECH_WEAP_MASS_DRIVER, /*13*/
    TECH_WEAP_MERCULITE_MISSILES, /*14*/
    TECH_WEAP_NEUTRON_BLASTER, /*15*/
    TECH_WEAP_ANTI_MATTER_BOMB, /*16*/
    TECH_WEAP_GRAVITON_BEAM, /*17*/
    TECH_WEAP_STINGER_MISSILES, /*18*/
    TECH_WEAP_HARD_BEAM, /*19*/
    TECH_WEAP_FUSION_BEAM, /*20*/
    TECH_WEAP_ION_STREAM_PROJECTOR, /*21*/
    TECH_WEAP_OMEGA_V_BOMB, /*22*/
    TECH_WEAP_ANTI_MATTER_TORPEDOES, /*23*/
    TECH_WEAP_FUSION_RIFLE, /*24*/
    TECH_WEAP_MEGABOLT_CANNON, /*25*/
    TECH_WEAP_PHASOR, /*26*/
    TECH_WEAP_SCATTER_PACK_VII_MISSILES, /*27*/
    TECH_WEAP_AUTO_BLASTER, /*28*/
    TECH_WEAP_PULSON_MISSILES, /*29*/
    TECH_WEAP_TACHYON_BEAM, /*30*/
    TECH_WEAP_HAND_PHASOR, /*31*/
    TECH_WEAP_GAUSS_AUTOCANNON, /*32*/
    TECH_WEAP_PARTICLE_BEAM, /*33*/
    TECH_WEAP_HERCULAR_MISSILES, /*34*/
    TECH_WEAP_PLASMA_CANNON, /*35*/
    TECH_WEAP_DEATH_RAY, /*36*/
    TECH_WEAP_DISRUPTOR, /*37*/
    TECH_WEAP_PULSE_PHASOR, /*38*/
    TECH_WEAP_NEUTRONIUM_BOMB, /*39*/
    TECH_WEAP_HELLFIRE_TORPEDOES, /*40*/
    TECH_WEAP_ZEON_MISSILES, /*41*/
    TECH_WEAP_PLASMA_RIFLE, /*42*/
    TECH_WEAP_PROTON_TORPEDOES, /*43*/
    TECH_WEAP_SCATTER_PACK_X_MISSILES, /*44*/
    TECH_WEAP_TRI_FOCUS_PLASMA_CANNON, /*45*/
    TECH_WEAP_STELLAR_CONVERTER, /*46*/
    TECH_WEAP_NEUTRON_STREAM_PROJECTOR, /*47*/
    TECH_WEAP_MAULER_DEVICE, /*48*/
    TECH_WEAP_UNUSED_48, /*49*/
    TECH_WEAP_PLASMA_TORPEDOES /*50*/
} tech_weap_t;

#endif
